import urllib.parse
import urllib3
import json
import requests
import pandas as pd
from importlib.resources import files, as_file


def escreve_descritores(
    query,
    esfera,
    orgao,
    uo,
    funcao,
    sub_funcao,
    programa,
    acao,
    subtitulo,
    plano_orcamentario,
    fonte,
    categoria_economica,
    gnd,
    modalidade_aplicacao,
    id_uso,
    resultado_primario,
    elemento_despesa,
    inclui_descricoes,
):
    """Adiciona descritores de dimensão à query SPARQL."""

    if esfera is not False:
        query += " ?codEsfera"
        if inclui_descricoes:
            query += " ?descEsfera"
    if orgao is not False:
        query += " ?codOrgao"
        if inclui_descricoes:
            query += " ?descOrgao"
    if uo is not False:
        query += " ?codUO"
        if inclui_descricoes:
            query += " ?descUO"
    if funcao is not False:
        query += " ?codFuncao"
        if inclui_descricoes:
            query += " ?descFuncao"
    if sub_funcao is not False:
        query += " ?codSubfuncao"
        if inclui_descricoes:
            query += " ?descSubfuncao"
    if programa is not False:
        query += " ?codPrograma"
        if inclui_descricoes:
            query += " ?descPrograma"
    if acao is not False:
        query += " ?codAcao"
        if inclui_descricoes:
            query += " ?descAcao"
    if subtitulo is not False:
        query += " ?codSubtitulo"
        if inclui_descricoes:
            query += " ?descSubtitulo"
    if plano_orcamentario is not False:
        query += " ?codPlanoOrcamentario"
        if inclui_descricoes:
            query += " ?descPlanoOrcamentario"
    if fonte is not False:
        query += " ?codFonte"
        if inclui_descricoes:
            query += " ?descFonte"
    if categoria_economica is not False:
        query += " ?codCategoriaEconomica"
        if inclui_descricoes:
            query += " ?descCategoriaEconomica"
    if gnd is not False:
        query += " ?codGND"
        if inclui_descricoes:
            query += " ?descGND"
    if modalidade_aplicacao is not False:
        query += " ?codModalidadeAplicacao"
        if inclui_descricoes:
            query += " ?descModalidadeAplicacao"
    if id_uso is not False:
        query += " ?codIdUso"
        if inclui_descricoes:
            query += " ?descIdUso"
    if resultado_primario is not False:
        query += " ?codResultadoPrimario"
        if inclui_descricoes:
            query += " ?descResultadoPrimario"
    if elemento_despesa is not False:
        query += " ?codElementoDespesa"
        if inclui_descricoes:
            query += " ?descElementoDespesa"

    return query


def constroi_siop_query_detalhe_anual(
    exercicio,
    esfera,
    orgao,
    uo,
    funcao,
    sub_funcao,
    programa,
    acao,
    plano_orcamentario,
    subtitulo,
    categoria_economica,
    gnd,
    modalidade_aplicacao,
    elemento_despesa,
    fonte,
    id_uso,
    resultado_primario,
    valor_ploa,
    valor_loa,
    valor_loa_mais_credito,
    valor_empenhado,
    valor_liquidado,
    valor_pago,
    inclui_descricoes,
):
    """Constrói query SPARQL detalhada anual."""

    query = "SELECT ?codExercicio"

    query = escreve_descritores(
        query=query,
        esfera=esfera,
        orgao=orgao,
        uo=uo,
        funcao=funcao,
        sub_funcao=sub_funcao,
        programa=programa,
        acao=acao,
        subtitulo=subtitulo,
        plano_orcamentario=plano_orcamentario,
        fonte=fonte,
        categoria_economica=categoria_economica,
        gnd=gnd,
        modalidade_aplicacao=modalidade_aplicacao,
        id_uso=id_uso,
        resultado_primario=resultado_primario,
        elemento_despesa=elemento_despesa,
        inclui_descricoes=inclui_descricoes,
    )

    if valor_ploa:
        query += " (sum(?val1) as ?ploa)"
    if valor_loa:
        query += " (sum(?val2) as ?loa)"
    if valor_loa_mais_credito:
        query += " (sum(?val3) as ?loa_mais_credito)"
    if valor_empenhado:
        query += " (sum(?val4) as ?empenhado)"
    if valor_liquidado:
        query += " (sum(?val5) as ?liquidado)"
    if valor_pago:
        query += " (sum(?val6) as ?pago)"

    query += f" WHERE {{ GRAPH <http://orcamento.dados.gov.br/{exercicio}/> {{"
    query += """
        ?i                     loa:temExercicio      ?exercicio .
        ?exercicio             loa:dataUltimaAtualizacao  ?data .
        ?exercicio             loa:identificador     ?codExercicio ."""


    if esfera is not False:
        query += """
        ?i                     loa:temEsfera         ?esfera .
        ?esfera                loa:codigo            ?codEsfera ."""
        if inclui_descricoes:
            query += " ?esfera                rdfs:label            ?descEsfera ."

    if orgao is not False:
        query += """
        ?UO                    loa:temOrgao          ?orgao .
        ?orgao                 loa:codigo            ?codOrgao ."""
        if inclui_descricoes:
            query += " ?orgao                 rdfs:label            ?descOrgao ."

    if uo is not False:
        query += """
        ?i                     loa:temUnidadeOrcamentaria  ?UO .
        ?UO                    loa:codigo            ?codUO ."""
        if inclui_descricoes:
            query += " ?UO                    rdfs:label            ?descUO ."

    if funcao is not False:
        query += """
        ?i                     loa:temFuncao         ?funcao .
        ?funcao                loa:codigo            ?codFuncao ."""
        if inclui_descricoes:
            query += " ?funcao                rdfs:label            ?descFuncao ."

    if sub_funcao is not False:
        query += """
        ?i                     loa:temSubfuncao      ?sub_funcao .
        ?sub_funcao             loa:codigo            ?codSubfuncao ."""
        if inclui_descricoes:
            query += " ?sub_funcao             rdfs:label            ?descSubfuncao ."

    if programa is not False:
        query += """
        ?i                     loa:temPrograma       ?programa .
        ?programa              loa:codigo            ?codPrograma ."""
        if inclui_descricoes:
            query += " ?programa              rdfs:label            ?descPrograma ."

    if acao is not False:
        query += """
        ?i                     loa:temAcao           ?acao .
        ?acao                  loa:codigo            ?codAcao ."""
        if inclui_descricoes:
            query += " ?acao                  rdfs:label            ?descAcao ."

    if subtitulo is not False:
        query += """
        ?i                     loa:temSubtitulo      ?subtitulo .
        ?subtitulo             loa:codigo            ?codSubtitulo ."""
        if inclui_descricoes:
            query += " ?subtitulo             rdfs:label            ?descSubtitulo ."

    if plano_orcamentario is not False:
        query += """
        ?i                     loa:temPlanoOrcamentario  ?plano_orcamentario .
        ?plano_orcamentario     loa:codigo            ?codPlanoOrcamentario ."""
        if inclui_descricoes:
            query += " ?plano_orcamentario     rdfs:label            ?descPlanoOrcamentario ."

    if fonte is not False:
        query += """
        ?i                     loa:temFonteRecursos  ?fonte .
        ?fonte                 loa:codigo            ?codFonte ."""
        if inclui_descricoes:
            query += " ?fonte                 rdfs:label            ?descFonte ."

    if categoria_economica is not False:
        query += """
        ?i                     loa:temCategoriaEconomica  ?categoria_economica .
        ?categoria_economica    loa:codigo            ?codCategoriaEconomica ."""
        if inclui_descricoes:
            query += " ?categoria_economica    rdfs:label            ?descCategoriaEconomica ."

    if gnd is not False:
        query += """
        ?i                     loa:temGND            ?gnd .
        ?gnd                   loa:codigo            ?codGND ."""
        if inclui_descricoes:
            query += " ?gnd                   rdfs:label            ?descGND ."

    if modalidade_aplicacao is not False:
        query += """
        ?i                     loa:temModalidadeAplicacao  ?modalidade_aplicacao .
        ?modalidade_aplicacao   loa:codigo            ?codModalidadeAplicacao ."""
        if inclui_descricoes:
            query += " ?modalidade_aplicacao   rdfs:label            ?descModalidadeAplicacao ."

    if id_uso is not False:
        query += """
        ?i                     loa:temIdentificadorUso  ?id_uso .
        ?id_uso                 loa:codigo            ?codIdUso ."""
        if inclui_descricoes:
            query += " OPTIONAL  {?id_uso      rdfs:label            ?descIdUso }."

    if resultado_primario is not False:
        query += """
        ?i                     loa:temResultadoPrimario  ?resultado_primario .
        ?resultado_primario     loa:codigo            ?codResultadoPrimario ."""
        if inclui_descricoes:
            query += " OPTIONAL {?resultado_primario rdfs:label ?descResultadoPrimario }."

    if elemento_despesa is not False:
        query += """
        ?i                     loa:temElementoDespesa  ?elemento_despesa .
        ?elemento_despesa       loa:codigo            ?codElementoDespesa ."""
        if inclui_descricoes:
            query += " ?elemento_despesa       rdfs:label            ?descElementoDespesa ."

    # FILTERS
    if isinstance(esfera, str):
        query += f' ?esfera                 loa:codigo "{esfera}" .'
    if isinstance(orgao, str):
        query += f' ?orgao                  loa:codigo "{orgao}" .'
    if type(uo) in (str, int):
        query += f' ?UO                     loa:codigo "{uo}" .'
    if isinstance(funcao, str):
        query += f' ?funcao                 loa:codigo  "{funcao}" .'
    if isinstance(sub_funcao, str):
        query += f' ?sub_funcao              loa:codigo  "{sub_funcao}" .'
    if isinstance(programa, str):
        query += f' ?programa               loa:codigo  "{programa}" .'
    if isinstance(acao, str):
        query += f' ?acao                   loa:codigo  "{acao}" .'
    if isinstance(subtitulo, str):
        query += f' ?subtitulo              loa:codigo  "{subtitulo}" .'
    if isinstance(plano_orcamentario, str):
        query += f' ?plano_orcamentario      loa:codigo  "{plano_orcamentario}" .'
    if isinstance(fonte, str):
        query += f' ?fonte                  loa:codigo  "{fonte}" .'
    if isinstance(categoria_economica, str):
        query += f' ?categoria_economica     loa:codigo  "{categoria_economica}" .'
    if isinstance(gnd, str):
        query += f' ?gnd                    loa:codigo  "{gnd}" .'
    if isinstance(modalidade_aplicacao, str):
        query += f' ?modalidade_aplicacao    loa:codigo  "{modalidade_aplicacao}" .'
    if isinstance(id_uso, str):
        query += f' ?id_uso                  loa:codigo  "{id_uso}" .'
    if type(resultado_primario) in (str, int):
        query += f' ?resultado_primario      loa:codigo  "{resultado_primario}" .'
    if isinstance(elemento_despesa, str):
        query += f' ?elemento_despesa        loa:codigo  "{elemento_despesa}" .'

    # Valores
    if valor_ploa:
        query += " ?i loa:valorProjetoLei      ?val1 ."
    if valor_loa:
        query += " ?i loa:valorDotacaoInicial ?val2 ."
    if valor_loa_mais_credito:
        query += " ?i loa:valorLeiMaisCredito ?val3 ."
    if valor_empenhado:
        query += " ?i loa:valorEmpenhado       ?val4 ."
    if valor_liquidado:
        query += " ?i loa:valorLiquidado       ?val5 ."
    if valor_pago:
        query += " ?i loa:valorPago            ?val6 ."

    query += " } } GROUP BY ?codExercicio"

    query = escreve_descritores(
        query=query,
        esfera=esfera,
        orgao=orgao,
        uo=uo,
        funcao=funcao,
        sub_funcao=sub_funcao,
        programa=programa,
        acao=acao,
        subtitulo=subtitulo,
        plano_orcamentario=plano_orcamentario,
        fonte=fonte,
        categoria_economica=categoria_economica,
        gnd=gnd,
        modalidade_aplicacao=modalidade_aplicacao,
        id_uso=id_uso,
        resultado_primario=resultado_primario,
        elemento_despesa=elemento_despesa,
        inclui_descricoes=inclui_descricoes,
    )

    return query


def constroi_siop_query_dimensoes(exercicio, dimensao):
    """Constroi a query SPARQL para listar membros de uma dimensão específica."""
    dimensoes_df = carregar_dataframe_dimensoes()
    var = dimensoes_df.loc[dimensoes_df['param'] == dimensao, 'var'].iloc[0]

    query = (
        "SELECT ?cod ?desc WHERE { "
        f"GRAPH <http://orcamento.dados.gov.br/{exercicio}/> {{ "
        f"?i loa:{var} ?dimensao . "
        "?dimensao loa:codigo ?cod . "
        "?dimensao rdfs:label ?desc . "
        "} } "
        "GROUP BY ?cod ?desc ORDER BY ?cod"
    )
    return query


def forma_url_siop(query, timeout=0):
    query = ' '.join(query.split())
    query_encoded = urllib.parse.quote(query, safe='')
    url = (
        "https://www1.siop.planejamento.gov.br/sparql/?default-graph-uri=&query="
        + query_encoded
        + "&format=application%2Fsparql-results%2Bjson"
        + f"&timeout={timeout}&debug=on"
    )
    return url


def download_siop(url_completo, ignore_secure_certificate):
    """Baixa dados do SIOP, com ou sem verificação SSL."""

    response = None

    if not ignore_secure_certificate:
        try:
            res = files('orcamentobr.certs') / 'fullchainsiop.pem'
            with as_file(res) as cert_path:
                response = requests.get(url_completo, verify=str(cert_path))

        except requests.exceptions.SSLError as e:
            raise Exception(f"Error using the certificate: {e}")
        except requests.exceptions.RequestException as e:
            raise Exception(f"Error accessing endpoint with certificate: {e}")
        except FileNotFoundError as e:
            raise Exception(f"Certificate file error: {e}")
        except Exception as e:
            raise Exception(f"Unexpected error during secure request: {e}")
    else:
        try:
            urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)
            response = requests.get(url_completo, verify=False)
        except requests.exceptions.RequestException as e:
            raise Exception(f"Error accessing endpoint (SSL ignored): {e}")
        except Exception as e:
            raise Exception(f"Unexpected error during insecure request: {e}")

    if response is None:
        raise Exception("Failed to get a response from the URL.")

    if response.status_code != 200:
        conteudo_erro = response.text.split("\n")[0] if response.text else "No error content."
        raise Exception(
            f"Error in downloading from endpoint. Code: {response.status_code}\n"
            f"{conteudo_erro}"
        )

    text_content = response.text.strip()

    if text_content.startswith("{") or text_content.startswith("["):
        try:
            json_data = json.loads(text_content)
            return json_data
        except json.JSONDecodeError as e:
            raise Exception(
                f"Failed to decode JSON: {e}. Content: {text_content[:200]}..."
            )
    else:
        raise Exception(
            f"Response is not JSON or does not start with '{{' or '['. "
            f"Content preview: {text_content[:100]}..."
        )


def json_para_dataframe(json_data):
    """Converte JSON retornado pelo SIOP em DataFrame do pandas."""
    df = pd.DataFrame(
        [
            {k: v['value'] for k, v in item.items()}
            for item in json_data['results']['bindings']
        ]
    )
    return df


def carregar_dataframe_dimensoes():
    dimensoes = pd.DataFrame(
        {
            'param': [
                "exercicio", "esfera", "orgao", "uo", "funcao", "sub_funcao",
                "programa", "acao", "plano_orcamentario", "subtitulo",
                "categoria_economica", "gnd", "modalidade_aplicacao",
                "elemento_despesa", "fonte", "id_uso", "resultado_primario"
            ],
            'var': [
                "temExercicio", "temEsfera", "temOrgao", "temUnidadeOrcamentaria",
                "temFuncao", "temSubfuncao", "temPrograma", "temAcao",
                "temPlanoOrcamentario", "temSubtitulo", "temCategoriaEconomica",
                "temGND", "temModalidadeAplicacao", "temElementoDespesa",
                "temFonteRecursos", "temIdentificadorUso", "temResultadoPrimario"
            ]
        }
    )
    return dimensoes



from orcamentobr import (
    quais_membros,
    constroi_url_siconfi,
    verifica_params,
    detailed_expenditure
)

def test_dummy():
    assert True

def test_quais_membros_runs():
    # Test only that the function runs and returns a DataFrame-like object
    try:
        result = quais_membros(2023, dimensao="Funcao")
        assert hasattr(result, "columns")
    except Exception:
        assert True  # Accepts failure if no internet or API changes

def test_constroi_url_siconfi():
    url = constroi_url_siconfi("entes")
    assert isinstance(url, str)
    assert "https://apidatalake.tesouro.gov.br/ords/siconfi/tt//entes" in url

def test_verifica_params_valid():
    params = {"id_ente": 1, "an_referencia": 2020, "me_referencia": 12, "co_tipo_matriz": "MSCE", "classe_conta": 6, "id_tv": "period_change"}
    result = verifica_params("msc_orcamentaria", params)
    assert result
    
    
    
def test_detailed_expenditure():
    # Exemplo de uso
    df = detailed_expenditure(
        exercicio=2023, Esfera=False, Orgao=False, UO=False, Funcao=False, Subfuncao=False,
        Programa=False, Acao=False, PlanoOrcamentario=False, Subtitulo=False,
        CategoriaEconomica=False, GND=False, ModalidadeAplicacao=False, ElementoDespesa=False,
        Fonte=False, IdUso=False, ResultadoPrimario=False,
        valorPLOA=True, valorLOA=True, valorLOAmaisCredito=True, valorEmpenhado=True, valorLiquidado=True, valorPago=True,
        incluiDescricoes=True, detalheMaximo=False, ignoreSecureCertificate=False, timeout=0, print_url=True
    )

    print(df)
    assert True



from datetime import datetime
from .siop_utils import (
    constroi_siop_query_dimensoes,
    forma_url_siop,
    carregar_dataframe_dimensoes,
    download_siop,
    json_para_dataframe,
)


def quais_membros(
    exercicio: int | None = None,
    dimensao: str | None = None,
    ignore_secure_certificate: bool = False,
):
    """Lista os membros de uma dimensão do Orçamento Federal para um dado ano.

    Parameters
    ----------
    exercicio : int
        O ano a que se refere a extração.
    dimensao : str
        A dimensão do orçamento a ser listada.
        Valores válidos: "Esfera", "Orgao", "UO", "Funcao", etc.
    ignore_secure_certificate : bool, optional
        Se True, o download dos dados do SIOP ignora o certificado de
        segurança. O padrão é False.

    Returns
    -------
    pandas.DataFrame
        Um DataFrame com os membros da dimensão, contendo tanto o código
        quanto a descrição de cada membro.

    Examples
    --------
    >>> # Listar todos os membros da dimensão 'Funcao' para o ano de 2023
    >>> quais_membros(exercicio=2023, dimensao="Funcao")

    >>> # Listar membros de outra dimensão, ignorando o certificado de segurança
    >>> quais_membros(
    ...     exercicio=2024, dimensao="Orgao", ignore_secure_certificate=True
    ... )
    """
    if exercicio is None:
        exercicio = datetime.now().year - 1

    dimensoes_df = carregar_dataframe_dimensoes()

    if not isinstance(dimensao, str):
        raise ValueError(
            "Erro: 'dimensao' deve ser uma string!\n"
            f"Valores válidos: {dimensoes_df['param'][1:17].tolist()}"
        )

    if dimensao not in dimensoes_df['param'][1:17].values:
        raise ValueError(
            "Erro: 'dimensao' não contém um valor válido!\n"
            f"Valores válidos: {', '.join(dimensoes_df['param'][1:17])}"
        )

    query = constroi_siop_query_dimensoes(exercicio=exercicio, dimensao=dimensao)
    url = forma_url_siop(query)
    dados = download_siop(url, ignore_secure_certificate)
    df = json_para_dataframe(dados)
    df.columns = [f"{dimensao}_cod", f"{dimensao}_desc"]

    return df


def which_members(
    exercicio: int | None = None,
    dimensao: str | None = None,
    ignore_secure_certificate: bool = False,
):
    """Lists the members in a Brazilian Federal Budget dimension.

    This function is a wrapper for `quais_membros`, providing an English
    interface.

    Parameters
    ----------
    exercicio : int
        The year to which the extracted data refers.
    dimensao : str
        The budget dimension to be listed.
        Valid values: "Esfera", "Orgao", "UO", "Funcao", etc.
    ignore_secure_certificate : bool, optional
        If True, the SIOP data download ignores the secure certificate.
        Default is False.

    Returns
    -------
    pandas.DataFrame
        A DataFrame containing the dimension members, including both the
        code and description for each member.

    Examples
    --------
    >>> # List all members of the 'Funcao' dimension for the year 2023
    >>> which_members(exercicio=2023, dimensao="Funcao")

    >>> # List members of another dimension, ignoring the secure certificate
    >>> which_members(
    ...     exercicio=2024, dimensao="Orgao", ignore_secure_certificate=True
    ... )
    """
    return quais_membros(
        exercicio=exercicio,
        dimensao=dimensao,
        ignore_secure_certificate=ignore_secure_certificate,
    )




from datetime import datetime
import pandas as pd

from .siop_utils import (
    forma_url_siop,
    carregar_dataframe_dimensoes,
    download_siop,
    json_para_dataframe,
    constroi_siop_query_detalhe_anual,
)


def despesa_detalhada(
    exercicio: int | None = None,
    esfera: bool | int | str = False,
    orgao: bool | int | str = False,
    uo: bool | int | str = False,
    funcao: bool | int | str = False,
    sub_funcao: bool | int | str = False,
    programa: bool | int | str = False,
    acao: bool | int | str = False,
    plano_orcamentario: bool | int | str = False,
    subtitulo: bool | int | str = False,
    categoria_economica: bool | int | str = False,
    gnd: bool | int | str = False,
    modalidade_aplicacao: bool | int | str = False,
    elemento_despesa: bool | int | str = False,
    fonte: bool | int | str = False,
    id_uso: bool | int | str = False,
    resultado_primario: bool | int | str = False,
    valor_ploa: bool = True,
    valor_loa: bool = True,
    valor_loa_mais_credito: bool = True,
    valor_empenhado: bool = True,
    valor_liquidado: bool = True,
    valor_pago: bool = True,
    inclui_descricoes: bool = True,
    detalhe_maximo: bool = False,
    ignore_secure_certificate: bool = False,
    timeout: int = 0,
    print_url: bool = False,
):
    """
    Download expenditure data from the Brazilian federal budget.

    Queries the SIOP API to retrieve expenditure data for a given fiscal year,
    with options to disaggregate by various qualitative and quantitative
    dimensions and to include selected financial metrics. Dimension parameters
    may be:
    - `bool`: `True` for full breakdown, `False` for aggregation
    - `int` or `str`: a specific code or identifier to filter that dimension

    Parameters
    ----------
    exercicio : int or None, default None
        Fiscal year to query. If None, uses the package’s internal `.last_year()`.
    esfera, orgao, uo, funcao, sub_funcao, programa, acao,
    plano_orcamentario, subtitulo, categoria_economica, gnd,
    modalidade_aplicacao, elemento_despesa, fonte, id_uso,
    resultado_primario : bool, int, or str, optional
        Dimension flags or filters (default False).
        - If `bool`: `True` to break down by this dimension, `False` to aggregate.
        - If `int` or `str`: filter results to the given code only.

        - esfera: Esfera Orçamentária
        - orgao: Órgão
        - uo: Unidade Orçamentária
        - funcao: Função
        - sub_funcao: Subfunção
        - programa: Programa
        - acao: Ação
        - plano_orcamentario: Plano Orçamentário
        - subtitulo: Subtítulo
        - categoria_economica: Categoria Econômica
        - gnd: Grupo Natureza da Despesa
        - modalidade_aplicacao: Modalidade de Aplicação
        - elemento_despesa: Elemento da Despesa
        - fonte: Fonte de Recursos
        - id_uso: Contrapartida (Identificador de uso)
        - resultado_primario: Identificador de resultado primário

    valor_ploa, valor_loa, valor_loa_mais_credito,
    valor_empenhado, valor_liquidado, valor_pago : bool, optional
        Which monetary fields to include (default True).

    inclui_descricoes : bool, optional
        Include descriptive labels for each selected dimension (default True).

    detalhe_maximo : bool, optional
        Override all other flags and disaggregate by all dimensions (default False).

    ignore_secure_certificate : bool, optional
        Skip SSL certificate validation if True (default False).

    timeout : int, optional
        Request timeout in milliseconds (default 0). Values <1000 are ignored.

    print_url : bool, optional
        Print the constructed API URL before request if True (default False).

    Returns
    -------
    pandas.DataFrame
        DataFrame with expenditure figures. Columns depend on flags provided.

    References
    ----------
    Manual Técnico do Orçamento, SIOP:
    https://www1.siop.planejamento.gov.br/mto/lib/exe/fetch.php/mto2025:mto2025.pdf
    """
    if exercicio is None:
        exercicio = datetime.now().year - 1

    args = dict(locals())
    args['exercicio'] = int(args['exercicio'])

    if args['detalhe_maximo']:
        dimensoes_df = carregar_dataframe_dimensoes()
        params = dimensoes_df['param'][1:17]  # dimensões de detalhe
        for param in params:
            args[param] = True

    args_para_query = {
        k: v for k, v in args.items()
        if k not in ['detalhe_maximo', 'ignore_secure_certificate', 'timeout', 'print_url']
    }

    query = constroi_siop_query_detalhe_anual(**args_para_query)
    xurl = forma_url_siop(query, timeout)

    if print_url:
        print(xurl)

    dados = download_siop(xurl, ignore_secure_certificate)
    df = json_para_dataframe(dados)

    # Transform numeric columns
    for col in ['ploa', 'loa', 'loa_mais_credito', 'empenhado', 'liquidado', 'pago']:
        if col in df.columns:
            df[col] = pd.to_numeric(df[col])

    # Rename columns
    novos_nomes = []
    for nome in df.columns:
        if nome == "codExercicio":
            novos_nomes.append("exercicio")
        elif nome.startswith("cod"):
            novos_nomes.append(f"{nome[3:]}_cod")
        elif nome.startswith("desc"):
            novos_nomes.append(f"{nome[4:]}_desc")
        else:
            novos_nomes.append(nome)
    df.columns = novos_nomes

    return df


def detailed_expenditure(
    exercicio: int | None = None,
    esfera: bool | int | str = False,
    orgao: bool | int | str = False,
    uo: bool | int | str = False,
    funcao: bool | int | str = False,
    sub_funcao: bool | int | str = False,
    programa: bool | int | str = False,
    acao: bool | int | str = False,
    plano_orcamentario: bool | int | str = False,
    subtitulo: bool | int | str = False,
    categoria_economica: bool | int | str = False,
    gnd: bool | int | str = False,
    modalidade_aplicacao: bool | int | str = False,
    elemento_despesa: bool | int | str = False,
    fonte: bool | int | str = False,
    id_uso: bool | int | str = False,
    resultado_primario: bool | int | str = False,
    valor_ploa: bool = True,
    valor_loa: bool = True,
    valor_loa_mais_credito: bool = True,
    valor_empenhado: bool = True,
    valor_liquidado: bool = True,
    valor_pago: bool = True,
    inclui_descricoes: bool = True,
    detalhe_maximo: bool = False,
    ignore_secure_certificate: bool = False,
    timeout: int = 0,
    print_url: bool = False,
):
    """
    Download expenditure data from the Brazilian federal budget.

    Queries the SIOP API to retrieve expenditure data for a given fiscal year,
    with options to disaggregate by various qualitative and quantitative
    dimensions and to include selected financial metrics. Dimension parameters
    may be:
    - `bool`: `True` for full breakdown, `False` for aggregation
    - `int` or `str`: a specific code or identifier to filter that dimension

    Parameters
    ----------
    exercicio : int or None, default None
        Fiscal year to query. If None, uses the package’s internal `.last_year()`.
    esfera, orgao, uo, funcao, sub_funcao, programa, acao,
    plano_orcamentario, subtitulo, categoria_economica, gnd,
    modalidade_aplicacao, elemento_despesa, fonte, id_uso,
    resultado_primario : bool, int, or str, optional
        Dimension flags or filters (default False).
        - If `bool`: `True` to break down by this dimension, `False` to aggregate.
        - If `int` or `str`: filter results to the given code only.

        - esfera: Esfera Orçamentária
        - orgao: Órgão
        - uo: Unidade Orçamentária
        - funcao: Função
        - sub_funcao: Subfunção
        - programa: Programa
        - acao: Ação
        - plano_orcamentario: Plano Orçamentário
        - subtitulo: Subtítulo
        - categoria_economica: Categoria Econômica
        - gnd: Grupo Natureza da Despesa
        - modalidade_aplicacao: Modalidade de Aplicação
        - elemento_despesa: Elemento da Despesa
        - fonte: Fonte de Recursos
        - id_uso: Contrapartida (Identificador de uso)
        - resultado_primario: Identificador de resultado primário

    valor_ploa, valor_loa, valor_loa_mais_credito,
    valor_empenhado, valor_liquidado, valor_pago : bool, optional
        Which monetary fields to include (default True).

    inclui_descricoes : bool, optional
        Include descriptive labels for each selected dimension (default True).

    detalhe_maximo : bool, optional
        Override all other flags and disaggregate by all dimensions (default False).

    ignore_secure_certificate : bool, optional
        Skip SSL certificate validation if True (default False).

    timeout : int, optional
        Request timeout in milliseconds (default 0). Values <1000 are ignored.

    print_url : bool, optional
        Print the constructed API URL before request if True (default False).

    Returns
    -------
    pandas.DataFrame
        DataFrame with expenditure figures. Columns depend on flags provided.

    References
    ----------
    Manual Técnico do Orçamento, SIOP:
    https://www1.siop.planejamento.gov.br/mto/lib/exe/fetch.php/mto2025:mto2025.pdf
    """
    return despesa_detalhada(
        exercicio=exercicio,
        esfera=esfera,
        orgao=orgao,
        uo=uo,
        funcao=funcao,
        sub_funcao=sub_funcao,
        programa=programa,
        acao=acao,
        plano_orcamentario=plano_orcamentario,
        subtitulo=subtitulo,
        categoria_economica=categoria_economica,
        gnd=gnd,
        modalidade_aplicacao=modalidade_aplicacao,
        elemento_despesa=elemento_despesa,
        fonte=fonte,
        id_uso=id_uso,
        resultado_primario=resultado_primario,
        valor_ploa=valor_ploa,
        valor_loa=valor_loa,
        valor_loa_mais_credito=valor_loa_mais_credito,
        valor_empenhado=valor_empenhado,
        valor_liquidado=valor_liquidado,
        valor_pago=valor_pago,
        inclui_descricoes=inclui_descricoes,
        detalhe_maximo=detalhe_maximo,
        ignore_secure_certificate=ignore_secure_certificate,
        timeout=timeout,
        print_url=print_url,
    )



# orcamentobr package
from .quais_dimensoes import quais_membros, which_members
from .despesa_detalhada import despesa_detalhada, detailed_expenditure


# Optionally, define _all_ for explicit export
_all_ = [
    "quais_membros",
    "which_members",
    "despesa_detalhada",
    "detailed_expenditure",
]